"use strict";
var __spreadArray = (this && this.__spreadArray) || function (to, from, pack) {
    if (pack || arguments.length === 2) for (var i = 0, l = from.length, ar; i < l; i++) {
        if (ar || !(i in from)) {
            if (!ar) ar = Array.prototype.slice.call(from, 0, i);
            ar[i] = from[i];
        }
    }
    return to.concat(ar || Array.prototype.slice.call(from));
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.makeBinaryString = exports.makeBinaryArray = exports.makeBinaryObject = exports.recordDataSection = exports.recordTextSection = exports.makeSymbolTable = void 0;
var constants_1 = require("@utils/constants");
var functions_1 = require("@utils/functions");
var makeSymbolTable = function (inputs) {
    /*
     * make symbol table from assembly file
     * using SYMBOL_TABLE in constants.ts
     *
     * 'dataSeg'에 data 저장
     * 'textSeg'에 text 저장
     *
     * .text
     * - indicates that following items are stored in the user text segment, typically instructions
     * - It always starts from 0x400000 (MEM_TEXT_START)
  
     * .data
     * - indicates that following data items are stored in the data segment
     * - It always starts from 0x10000000 (MEM_DATA_START)
     *
     * return
     * {
     *    dataSeg : dataSeg,
     *    textSeg : textSeg,
     *    dataSectionSize : dataSectionSize,
     *    textSectionSize : textSectionSize
     * }
     *
     * [USAGE EXAMPLE]
     * const {dataSeg, textSeg, dataSectionSize, textSectionSize} = makeSymbolTable(inputs);
    */
    (0, constants_1.resetSymbolTable)();
    var address = 0;
    var curSection = constants_1.section.MAX_SIZE;
    var dataSectionSize = 0;
    var textSectionSize = 0;
    var dataSeg = [];
    var textSeg = [];
    inputs.forEach(function (input) {
        var splited = input.split('\t').filter(function (s) { return s !== ''; }); // ex. ['array:', '.word', '3']
        var symbol = new constants_1.symbolT();
        if (splited[0] == '.data') {
            curSection = constants_1.section.DATA;
            address = constants_1.MEM_DATA_START;
            return;
        }
        else if (splited[0] == '.text') {
            curSection = constants_1.section.TEXT;
            address = constants_1.MEM_TEXT_START;
            return;
        }
        else if (curSection === constants_1.section.DATA) {
            if (splited.length === 2) {
                // ex. ['.word','123']
                dataSeg.push(splited[1]);
            }
            else {
                // ex. ['array:', '.word', '3']
                symbol.address = address;
                symbol.name = splited[0].replace(':', '');
                (0, functions_1.symbolTableAddEntry)(symbol);
                dataSeg.push(splited[2]);
            }
            dataSectionSize += constants_1.BYTES_PER_WORD;
        }
        else if (curSection === constants_1.section.TEXT) {
            if (splited.length === 1) {
                // ex. ['main:']
                symbol.name = splited[0].replace(':', '');
                symbol.address = address;
                (0, functions_1.symbolTableAddEntry)(symbol);
                return;
            }
            else {
                // ex. ['and', '$17, $17, $0']
                var name_1 = splited[0];
                textSeg.push(input); // ex. 'and	$17, $17, $0'
                if (name_1 === 'la') {
                    var targetSymbol = splited[1].split(' ')[1]; // ex. 'data1'
                    var targetAddress = (0, functions_1.toHexAndPad)(constants_1.SYMBOL_TABLE[targetSymbol]);
                    if (targetAddress.slice(4) !== '0000') {
                        textSectionSize += constants_1.BYTES_PER_WORD;
                        address += constants_1.BYTES_PER_WORD;
                    }
                }
            }
            textSectionSize += constants_1.BYTES_PER_WORD;
        }
        address += constants_1.BYTES_PER_WORD;
    });
    return { dataSeg: dataSeg, textSeg: textSeg, dataSectionSize: dataSectionSize, textSectionSize: textSectionSize };
};
exports.makeSymbolTable = makeSymbolTable;
function recordTextSection(textSeg) {
    /*
     * parameter로 textSeg를 받는다.
     * textSeg 있는 text들 한 줄 씩 체크해서 binaryText 리스트에 바이너리 문장으로 추가
     * 명령어 타입별(R, I, J)로 명령어 이름별로 묶어서 번역
     *
     *  binaryText 이라는 list에 명령어를 번역한 binary 문장을 한 줄씩 추가
     *  return binaryText
     *  binaryText: ['00000000000000000000000001011000', '00000000000000000000000000001100']
     */
    var curAddress = constants_1.MEM_TEXT_START;
    var instruct;
    var address;
    var rs;
    var rt;
    var rd;
    var imm;
    var shamt;
    var immReg;
    var binaryText = [];
    var mappingTable = [];
    var binaryInstructionCounter = 0;
    textSeg.forEach(function (text, i) {
        mappingTable.push([]);
        instruct = text.slice(1).replace(/ /g, '').split(/,|\t/);
        var opName = instruct[0];
        if (opName === 'la') {
            address = constants_1.SYMBOL_TABLE[instruct[2]];
            rt = (0, functions_1.numToBits)(Number(instruct[1].replace('$', '')), 5);
            imm = (0, functions_1.numToBits)(parseInt(address.toString(16).slice(0, 4), 16), 16);
            binaryText.push('001111' + '00000' + rt + imm);
            if (address.toString(16).slice(4, 8) !== '0000') {
                imm = (0, functions_1.numToBits)(parseInt(address.toString(16).slice(4, 8), 16), 16);
                binaryText.push('001101' + rt + rt + imm);
                curAddress += constants_1.BYTES_PER_WORD;
                // if two binary instructions are made by la instruction, then it should map 'current assembly instruction' to 'two binary instructions'
                mappingTable[i].push(binaryInstructionCounter);
                binaryInstructionCounter++;
            }
        }
        else if (opName === 'move') {
            //op = ADD op "000000"
            rs = (0, functions_1.numToBits)(Number(instruct[2].replace('$', '')), 5);
            rt = '00000';
            rd = (0, functions_1.numToBits)(Number(instruct[1].replace('$', '')), 5);
            shamt = '00000';
            binaryText.push('000000' + rs + rt + rd + shamt + '100000'); //funct = "100000"
        }
        else {
            var opInfo = constants_1.instList[opName];
            if (opInfo.type === 'R') {
                if (opInfo.name === 'sll' || opInfo.name === 'srl') {
                    rs = '00000';
                    rt = (0, functions_1.numToBits)(Number(instruct[2].replace('$', '')), 5);
                    rd = (0, functions_1.numToBits)(Number(instruct[1].replace('$', '')), 5);
                    shamt = (0, functions_1.numToBits)(Number(instruct[3]), 5);
                }
                else if (opInfo.name === 'jr') {
                    rs = (0, functions_1.numToBits)(Number(instruct[1].replace('$', '')), 5);
                    rt = '00000';
                    rd = '00000';
                    shamt = '00000';
                }
                else {
                    rs = (0, functions_1.numToBits)(Number(instruct[2].replace('$', '')), 5);
                    rt = (0, functions_1.numToBits)(Number(instruct[3].replace('$', '')), 5);
                    rd = (0, functions_1.numToBits)(Number(instruct[1].replace('$', '')), 5);
                    shamt = '00000';
                }
                binaryText.push(opInfo.op + rs + rt + rd + shamt + opInfo.funct);
            }
            else if (opInfo.type === 'I') {
                if (opInfo.name === 'lui') {
                    rt = (0, functions_1.numToBits)(Number(instruct[1].replace('$', '')), 5);
                    rs = '00000';
                    imm =
                        instruct[2].slice(0, 2) === '0x'
                            ? (0, functions_1.numToBits)(parseInt(instruct[2].slice(2), 16), 16)
                            : (0, functions_1.numToBits)(Number(instruct[2]), 16);
                }
                else if (opInfo.name === 'beq' || opInfo.name === 'bne') {
                    imm = (0, functions_1.numToBits)(Number((constants_1.SYMBOL_TABLE[instruct[3]] - (curAddress + 4)) / 4), 16);
                    rs = (0, functions_1.numToBits)(Number(instruct[1].replace('$', '')), 5);
                    rt = (0, functions_1.numToBits)(Number(instruct[2].replace('$', '')), 5);
                }
                else if (opInfo.name === 'lw' ||
                    opInfo.name === 'lhu' ||
                    opInfo.name === 'sw' ||
                    opInfo.name === 'sh') {
                    immReg = instruct[2].split('(')[1].split(')')[0];
                    rs = (0, functions_1.numToBits)(Number(immReg.replace('$', '')), 5);
                    rt = (0, functions_1.numToBits)(Number(instruct[1].replace('$', '')), 5);
                    imm = (0, functions_1.numToBits)(Number(instruct[2].split('(')[0]), 16);
                }
                else {
                    rs = (0, functions_1.numToBits)(Number(instruct[2].replace('$', '')), 5);
                    rt = (0, functions_1.numToBits)(Number(instruct[1].replace('$', '')), 5);
                    imm =
                        instruct[3].slice(0, 2) === '0x'
                            ? (0, functions_1.numToBits)(parseInt(instruct[3].slice(2), 16), 16)
                            : (0, functions_1.numToBits)(Number(instruct[3]), 16);
                }
                binaryText.push(opInfo.op + rs + rt + imm);
            }
            else if (opInfo.type === 'J') {
                address = Number(constants_1.SYMBOL_TABLE[instruct[1]]) / 4;
                binaryText.push(opInfo.op + (0, functions_1.numToBits)(address, 26));
            }
        }
        curAddress += constants_1.BYTES_PER_WORD;
        mappingTable[i].push(binaryInstructionCounter);
        binaryInstructionCounter++;
    });
    return [binaryText, mappingTable];
}
exports.recordTextSection = recordTextSection;
function recordDataSection(dataSeg) {
    /*
     * input값을 dataSeg를 받는다.
     * dataSeg에 있는 data들 한 줄 씩 체크해서 binaryData 리스트에 바이너리 문장으로 추가
     * data값을 그대로 binary 문자로 번역
     *
     *  binaryData 이라는 list에 명령어를 번역한 binary 문장을 한 줄씩 추가
     *  return binaryData
     *  ex) binaryData: ['00000010001000001000100000100100', '00000010010000001001000000100100']
     */
    var dataNum;
    var binaryData = [];
    for (var _i = 0, dataSeg_1 = dataSeg; _i < dataSeg_1.length; _i++) {
        var data = dataSeg_1[_i];
        dataNum =
            data.slice(0, 2) === '0x' ? parseInt(data.slice(2), 16) : Number(data);
        binaryData.push((0, functions_1.numToBits)(dataNum));
    }
    return binaryData;
}
exports.recordDataSection = recordDataSection;
function makeBinaryObject(inputs) {
    var _a = (0, exports.makeSymbolTable)(inputs), dataSeg = _a.dataSeg, textSeg = _a.textSeg, dataSectionSize = _a.dataSectionSize, textSectionSize = _a.textSectionSize;
    var _b = recordTextSection(textSeg), binaryText = _b[0], mappingTable = _b[1];
    var binaryData = recordDataSection(dataSeg);
    return {
        dataSectionSize: dataSectionSize,
        textSectionSize: textSectionSize,
        binaryText: binaryText,
        binaryData: binaryData,
        mappingTable: mappingTable,
        dataSeg: dataSeg,
        textSeg: textSeg,
    };
}
exports.makeBinaryObject = makeBinaryObject;
function makeBinaryArray(dataSectionSize, textSectionSize, binaryText, binaryData) {
    var output = __spreadArray(__spreadArray([
        (0, functions_1.numToBits)(textSectionSize, 32),
        (0, functions_1.numToBits)(dataSectionSize, 32)
    ], binaryText, true), binaryData, true);
    return output;
}
exports.makeBinaryArray = makeBinaryArray;
function makeBinaryString(dataSectionSize, textSectionSize, binaryText, binaryData) {
    /*
     * output에 text 문장 개수를 binary로 번역해서 추가
     * output에 data 개수를 binary로 번역해서 추가
     *
     */
    var binarySize = [
        (0, functions_1.numToBits)(textSectionSize, 32),
        (0, functions_1.numToBits)(dataSectionSize, 32),
    ];
    var output = '';
    binarySize.concat(binaryText, binaryData).map(function (binaryLine) {
        output += "".concat(binaryLine, "\n");
    });
    return output;
}
exports.makeBinaryString = makeBinaryString;
